package broker

import (
	"errors"
	"fmt"
	"log"
	"sync"

	"github.com/pivotal-cf/on-demand-service-broker-sdk/bosh"
	"github.com/pivotal-cf/on-demand-service-broker-sdk/serviceadapter"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
)

type Broker struct {
	boshClient     BoshClient
	cfClient       CloudFoundryClient
	serviceAdapter ServiceAdapterClient
	deployer       Deployer
	deploymentLock *sync.Mutex

	serviceOffering ServiceOffering

	logger *log.Logger

	// TODO is this feature ever going to be resurrected? If not, delete this flag and any code path that depends on it being true.
	// There is no non-test code path that sets this flag to true, hence all integration tests have been removed.
	PreDeleteDeploymentHookFeatureFlag bool
}

func New(boshClient BoshClient, cfClient CloudFoundryClient, serviceAdapter ServiceAdapterClient,
	deployer Deployer, serviceOffering ServiceOffering, logger *log.Logger) (*Broker, error) {

	plansToInstanceCounts, err := cfClient.CountInstancesOfPlans(serviceOffering.ID)
	if err != nil {
		return nil, err
	}

	for existingPlanID, existingPlanCount := range plansToInstanceCounts {
		if existingPlanCount > 0 && !serviceOffering.Plans.Contains(existingPlanID) {
			return nil, fmt.Errorf("You cannot change the plan_id of a plan that has existing service instances")
		}
	}

	return &Broker{
		boshClient:     boshClient,
		cfClient:       cfClient,
		serviceAdapter: serviceAdapter,
		deployer:       deployer,
		deploymentLock: &sync.Mutex{},

		serviceOffering: serviceOffering,

		logger: logger,
	}, nil
}

type OperationType string

const (
	OperationTypeCreate  = OperationType("Create")
	OperationTypeUpdate  = OperationType("Update")
	OperationTypeUpgrade = OperationType("Upgrade")
	OperationTypeDelete  = OperationType("Delete")
)

type OperationData struct {
	BoshTaskID    int
	OperationType OperationType
}

type InstanceState struct {
	PlanID              string
	OperationInProgress bool
}

type NotFoundError struct {
	error
}

func NewNotFoundError() NotFoundError {
	return NotFoundError{error: errors.New("resource was not found")}
}

func IsNotFound(err error) bool {
	switch err.(type) {
	case NotFoundError:
		return true
	default:
		return false
	}
}

type DeploymentNotFoundError struct {
	error
}

func NewDeploymentNotFoundError(e error) error {
	return DeploymentNotFoundError{e}
}

type OperationInProgressError struct {
	error
}

func NewOperationInProgressError(e error) error {
	return OperationInProgressError{e}
}

type ServiceAdapterCommandNotImplementedError struct {
	error
}

func NewServiceAdapterCommandNotImplementedError(e error) ServiceAdapterCommandNotImplementedError {
	return ServiceAdapterCommandNotImplementedError{e}
}

type DisplayableError struct {
	WrappedError error
}

func (e DisplayableError) Error() string {
	return e.WrappedError.Error()
}

type AdapterCommandError struct {
	messageForCLIUser  string
	messageForOperator string
}

func (e AdapterCommandError) Error() string {
	return e.messageForOperator
}

func (e AdapterCommandError) MessageForCLIUser() string {
	return e.messageForCLIUser
}

func NewAdapterCommandError(messageForCLIUser, messageForOperator string) AdapterCommandError {
	return AdapterCommandError{messageForCLIUser: messageForCLIUser, messageForOperator: messageForOperator}
}

const genericErrorMessage = "An internal error has occurred"

//go:generate counterfeiter -o fakes/fake_deployer.go . Deployer
type Deployer interface {
	Deploy(instanceID, planID, deploymentName string, operationType OperationType, arbitraryParams map[string]interface{}, oldManifest []byte, previousPlanID *string) (int, []byte, error)
}

//go:generate counterfeiter -o fakes/fake_service_adapter_client.go . ServiceAdapterClient
type ServiceAdapterClient interface {
	GenerateManifest(serviceReleases serviceadapter.ServiceDeployment, plan serviceadapter.Plan, requestParams map[string]interface{}, previousManifest []byte, previousPlan *serviceadapter.Plan) ([]byte, error)
	CreateBinding(bindingID string, deploymentTopology bosh.BoshVMs, manifest []byte, requestParams map[string]interface{}) (serviceadapter.Binding, error)
	DeleteBinding(bindingID string, deploymentTopology bosh.BoshVMs, manifest []byte, requestParams map[string]interface{}) error
	PreDeleteDeployment(instanceID string, deploymentTopology bosh.BoshVMs, manifest []byte) error
	GenerateDashboardUrl(instanceID string, plan serviceadapter.Plan, manifest []byte) (string, error)
}

//go:generate counterfeiter -o fakes/fake_bosh_client.go . BoshClient
type BoshClient interface {
	DeployRaw(manifest []byte) (int, error)
	GetTask(taskID int) (boshclient.BoshTask, error)
	GetTasks(deploymentName string) ([]boshclient.BoshTask, error)
	VMs(deploymentName string) (bosh.BoshVMs, error)
	DeleteDeployment(name string) (int, error)
	GetManifestRaw(name string) ([]byte, error)
}

//go:generate counterfeiter -o fakes/fake_cloud_foundry_client.go . CloudFoundryClient
type CloudFoundryClient interface {
	CountInstancesOfPlan(serviceOfferingID, planID string) (int, error)
	CountInstancesOfPlans(serviceOfferingID string) (instanceCountByPlanID map[string]int, err error)
	GetInstanceState(serviceInstanceGUID string) (InstanceState, error)
	GetInstancesOfServiceOffering(serviceOfferingID string) ([]string, error)
}
