package cloud_foundry_client

import (
	"fmt"
	"log"

	"github.com/pivotal-cf/on-demand-service-broker/authorizationheader"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
)

type Client struct {
	httpJsonClient
	url string
}

func New(url string, logger *log.Logger, authHeaderBuilder authorizationheader.AuthHeaderBuilder, trustedCertPEM []byte, disableTLSCertVerification bool) Client {
	httpClient := newWrappedHttpClient(logger, authHeaderBuilder, trustedCertPEM, disableTLSCertVerification)
	return Client{httpJsonClient: httpClient, url: url}
}

func (c Client) CountInstancesOfPlans(serviceId string) (map[string]int, error) {
	plans, err := c.getPlansForServiceID(serviceId)
	if err != nil {
		return map[string]int{}, err
	}

	output := map[string]int{}
	for _, plan := range plans {
		count, err := c.countServiceInstancesOfServicePlan(plan.ServicePlanEntity.ServiceInstancesUrl)
		if err != nil {
			return nil, err
		}
		output[plan.ServicePlanEntity.UniqueID] = count
	}

	return output, nil
}

func (c Client) GetInstanceState(serviceInstanceGUID string) (broker.InstanceState, error) {
	instance, err := c.getServiceInstance(serviceInstanceGUID)
	if err != nil {
		return broker.InstanceState{}, err
	}

	plan, err := c.getServicePlan(instance.Entity.ServicePlanURL)
	if err != nil {
		return broker.InstanceState{}, err
	}

	return broker.InstanceState{PlanID: plan.ServicePlanEntity.UniqueID, OperationInProgress: instance.Entity.LastOperation.State == "in progress"}, nil
}

func (c Client) CountInstancesOfPlan(serviceID, servicePlanID string) (int, error) {
	plans, err := c.getPlansForServiceID(serviceID)
	if err != nil {
		return 0, err
	}

	for _, plan := range plans {
		if plan.ServicePlanEntity.UniqueID == servicePlanID {
			count, err := c.countServiceInstancesOfServicePlan(plan.ServicePlanEntity.ServiceInstancesUrl)
			if err != nil {
				return 0, err
			}
			return count, nil
		}
	}

	return 0, fmt.Errorf("service plan %s not found for service %s", servicePlanID, serviceID)
}

func (c Client) GetInstancesOfServiceOffering(serviceOfferingID string) ([]string, error) {
	plans, err := c.getPlansForServiceID(serviceOfferingID)
	if err != nil {
		return nil, err
	}

	var instances []string
	for _, plan := range plans {
		var serviceInstancesResp serviceInstancesResponse
		if err := c.Get(fmt.Sprintf("%s/v2/service_plans/%s/service_instances", c.url, plan.Metadata.GUID), &serviceInstancesResp); err != nil {
			return nil, err
		}
		for _, instance := range serviceInstancesResp.ServiceInstances {
			instances = append(instances, instance.Metadata.GUID)
		}
	}
	return instances, nil
}

func (c Client) getPlansForServiceID(serviceID string) ([]ServicePlan, error) {
	requiredService, err := c.findServiceByUniqueID(serviceID)
	if err != nil {
		return nil, err
	}

	if requiredService == nil {
		return nil, nil
	}

	return c.listAllPlans(requiredService.ServiceEntity.ServicePlansUrl)
}

const DefaultPerPage = "100"

func (c Client) listServices(path string) (serviceResponse, error) {
	resp := serviceResponse{}
	return resp, c.Get(fmt.Sprintf("%s%s", c.url, path), &resp)
}

func (c Client) findServiceByUniqueID(uniqueID string) (*service, error) {
	path := "/v2/services?results-per-page=" + DefaultPerPage
	for path != "" {
		resp, err := c.listServices(path)
		if err != nil {
			return nil, err
		}
		if service := resp.Services.findByUniqueID(uniqueID); service != nil {
			return service, nil
		}
		path = resp.NextPath
	}

	return nil, nil
}

func (c Client) getServiceInstance(serviceInstanceGUID string) (serviceInstance, error) {
	path := fmt.Sprintf("/v2/service_instances/%s", serviceInstanceGUID)
	var instance serviceInstance
	err := c.Get(fmt.Sprintf("%s%s", c.url, path), &instance)
	return instance, err
}

func (c Client) getServicePlan(servicePlanPath string) (ServicePlan, error) {
	var plan ServicePlan
	err := c.Get(fmt.Sprintf("%s%s", c.url, servicePlanPath), &plan)
	return plan, err
}

func (c Client) listAllPlans(url string) ([]ServicePlan, error) {
	plans := []ServicePlan{}
	path := url + "?results-per-page=" + DefaultPerPage
	for path != "" {
		servicePlanResponse, err := c.listPlans(path)
		if err != nil {
			return nil, err
		}
		plans = append(plans, servicePlanResponse.ServicePlans...)
		path = servicePlanResponse.NextPath
	}

	return plans, nil
}
func (c Client) listPlans(path string) (ServicePlanResponse, error) {
	servicePlanResponse := ServicePlanResponse{}
	return servicePlanResponse, c.Get(fmt.Sprintf("%s%s", c.url, path), &servicePlanResponse)
}
func (c Client) countServiceInstancesOfServicePlan(path string) (int, error) {
	resp := serviceInstancesResponse{}
	err := c.Get(fmt.Sprintf("%s%s", c.url, path), &resp)
	if err != nil {
		return 0, err
	}
	return resp.TotalResults, nil
}
